from re import match
import asyncio, time


from .document import Document


async def match_data(data: Document or dict, query: dict) -> bool:
    """Match a document with a query.

    Args:
        data: The document to match.
        query: The query to match with.

    Example:
        >>> await match_data({"name": "Ashen"}, {"name": "Ashen"})
        True

    Note:
        All the operators might not work. Please report as bug.

        .. list-table::
            :header-rows: 1
            :widths: 20 20 20 20 20 20

            * - Comparison
              - Logical
              - Element
              - Evaluation
              - Geospatial
              - Array
            * - $eq
              - $and
              - $exists
              - $mod
              - $geoIntersects
              - $all
            * - $ne
              - $or
              - $type
              - $regex
              - $geoWithin
              - $elemMatch
            * - $gt
              - $not
              -
              - $text
              - $near
              - $slice
            * - $gte
              - $nor
              -
              - $where
              - $nearSphere
              -
            * - $lt
              - $elemMatch
              -
              -
              -
              -
            * - $lte
              -
              -
              -
              -
              -
            * - $in
              -
              -
              -
              -
              -
            * - $nin
              -
              -
              -
              -
              -

    """

    comparison_operators = {
        "$eq": lambda x, y: x == y,
        "$ne": lambda x, y: x != y,
        "$gt": lambda x, y: x > y,
        "$gte": lambda x, y: x >= y,
        "$lt": lambda x, y: x < y,
        "$lte": lambda x, y: x <= y,
        "$in": lambda x, y: x in y,
        "$nin": lambda x, y: x not in y,
    }

    async def and_operator(x, y):
        for i in y:
            if not await match_data(x, i):
                return False
        return True

    async def or_operator(x, y):
        for i in y:
            if await match_data(x, i):
                return True
        return False

    async def not_operator(x, y):
        return not await match_data(x, y)

    async def nor_operator(x, y):
        for i in y:
            if await match_data(x, i):
                return False
        return True

    logical_operators = {
        "$and": and_operator,
        "$or": or_operator,
        "$not": not_operator,
        "$nor": nor_operator,
    }

    element_operators = {
        "$exists": lambda x, y: y in x,
        "$type": lambda x, y: type(x) == y,
    }
    evalutation_operators = {
        "$mod": lambda x, y: x % y[0] == y[1],
        "$regex": lambda x, y: y.match(x),
        "$text": lambda x, y: y in x,
        "$where": lambda x, y: y(x),
    }
    geospatial_operators = {
        "$geoIntersects": lambda x, y: x in y,
        "$geoWithin": lambda x, y: x in y,
        "$near": lambda x, y: x in y,
        "$nearSphere": lambda x, y: x in y,
    }

    async def elemMatch_operator(x, y):
        for i in x:
            if await match_data(i, y):
                return True
        return False

    array_operators = {
        "$all": lambda x, y: all([i in x for i in y]),
        "$elemMatch": elemMatch_operator,
        "$size": lambda x, y: len(x) == y,
    }

    query_operators = {
        **comparison_operators,
        **logical_operators,
        **element_operators,
        **evalutation_operators,
        **geospatial_operators,
        **array_operators,
    }

    all_matched = True

    for key, value in query.items():
        # If there is a dot in the key, it means that we have to go deeper
        if "." in key:
            key, subkey = key.split(".", 1)
            try:
                result = await match_data(data=data[key], query={subkey: value})
                all_matched = all_matched and result
            except:
                continue
        # If the key starts with $, it means that it is an operator
        elif key.startswith("$") and key in query_operators:
            # Check if function is async
            if asyncio.iscoroutinefunction(query_operators[key]):
                # If it is async, we have to await it
                result = await query_operators[key](data, value)
            else:
                # If it is not async, we just call it
                result = query_operators[key](data, value)
            # Update all_matched
            all_matched = all_matched and result
        # If the value is a dict, it means that we have to go deeper
        elif isinstance(value, dict):
            try:
                result = await match_data(data[key], value)
            except:
                result = False
            all_matched = all_matched and result
        else:
            result = data[key] == value
            all_matched = all_matched and result

    return all_matched


async def update_data(document: Document or dict, update: dict) -> Document:
    """
    Update a document with the given update query.

    Args:
        document (Document): The document to update.
        update (dict): The update query.

    Example:
        >>> document = {"name": "John", "age": 20}
        >>> update = {"$set": {"age": 21}}
        >>> await update_data(document, update)
        {"name": "John", "age": 21}

    Note:
        All the operators might not work. Please report as bug.

        .. list-table::
            :header-rows: 1
            :widths: 25 25 25

            * - Field Operators
              - Array Operators
              - Modification Operators
            * - $currentDate ✔
              - $
              - $each
            * - $inc ✔
              - $addToSet ✔
              - $position
            * - $min ✔
              - $pop ✔
              - $slice
            * - $max ✔
              - $pull ✔
              - $sort
            * - $mul ✔
              - $push ✔
              -
            * - $rename ✔
              - $pullAll ✔
              -
            * - $set ✔
              -
              -
            * - $setOnInsert ✔
              -
              -
            * - $unset ✔
              -
              -
    """
    field_operators = {
        "$currentDate": lambda parent_key, key, value: parent_key.__setitem__(
            key, int(time.time())
        )
        or parent_key,
        "$inc": lambda parent_key, key, value: parent_key.__setitem__(
            key, parent_key[key] + value
        )
        or parent_key,
        "$min": lambda parent_key, key, value: parent_key.__setitem__(
            key, min(value, parent_key.get(key, float("inf")))
        )
        or parent_key,
        "$max": lambda parent_key, key, value: parent_key.__setitem__(
            key, max(value, parent_key.get(key, float("-inf")))
        )
        or parent_key,
        "$mul": lambda parent_key, key, value: parent_key.__setitem__(
            key, parent_key[key] * value
        )
        or parent_key,
        "$rename": lambda parent_key, key, value: parent_key.__setitem__(
            value, parent_key.pop(key)
        )
        or parent_key,
        "$set": lambda parent_key, key, value: parent_key.__setitem__(key, value)
        or parent_key,
        # setOnInsert inserts the value only if that didn't exist before
        "$setOnInsert": lambda parent_key, key, value: parent_key.__setitem__(
            key, value
        )
        if key not in parent_key
        else parent_key,
        "$unset": lambda parent_key, key, value: parent_key.__delitem__(key)
        if key in parent_key
        else None,
    }
    array_operators = {
        # "$" Acts as a placeholder to update the first element that matches the query condition.
        "$": lambda parent_key, key, query, update: "THIS IS NOT IMPLEMENTED YET",
        "$[]": lambda parent_key, key, query, update: "THIS IS NOT IMPLEMENTED YET",
        "$[<identifier>]": lambda parent_key, key, query, update: "THIS IS NOT IMPLEMENTED YET",
        "$addToSet": lambda parent_key, key, value: parent_key.__setitem__(
            key, parent_key[key] + [value]
        )
        if value not in parent_key[key]
        else parent_key,
        "$pop": lambda parent_key, key, value: parent_key.__setitem__(
            key, parent_key[key][:-1]
        )
        if value == -1
        else parent_key.__setitem__(key, parent_key[key][1:])
        if value == 1
        else parent_key,
        "$pull": lambda parent_key, key, value: parent_key.__setitem__(
            key,
            [
                item
                for item in parent_key[key]
                if not (
                    isinstance(item, dict)
                    and all(k in item and item[k] == v for k, v in value.items())
                )
            ],
        )
        if isinstance(value, dict)
        else parent_key.__setitem__(
            key,
            [item for item in parent_key[key] if item not in value],
        ),
        "$push": lambda parent_key, key, value: parent_key.__setitem__(
            key, parent_key[key].extend(value)
        )
        if key in parent_key and hasattr(value, "__iter__")
        else parent_key.__setitem__(key, value if isinstance(value, list) else [value]),
        "$pullAll": lambda parent_key, key, value: parent_key.__setitem__(
            key, [item for item in parent_key[key] if item not in value]
        )
        if isinstance(value, list)
        else parent_key,
    }

    modification_operators = {
        "$each": lambda x, y: y,
        "$position": lambda x, y: y,
        "$slice": lambda x, y: y,
        "$sort": lambda x, y: y,
    }

    update_operators = {
        **field_operators,
        **array_operators,
        **modification_operators,
    }

    for operator, data in update.items():
        for key, value in data.items():
            temp = document
            keys = key.split(".")
            for key in keys[:-1]:
                if "[" in key:
                    key, index = key.split("[")
                    index = int(index[:-1])
                    temp = temp[key][index]
                else:
                    temp = temp.setdefault(key, {})

            if operator in update_operators:
                update_operators[operator](temp, keys[-1], value)

    await document.save()
    return document



import aiofiles.os as aios
from typing import Generator

from .database import Database
from .exception import *


# Client Class
class AshenDB:
    """
    Entry point for AshenDB. This class is used to create, get, and delete databases.

    """

    base: str = ".ashendb/"

    @classmethod
    async def get_db(cls, db_name: str) -> Database:
        """Get a single database from database name.

        Args:
            db_name: The name of the database.

        Raises:
            NotFound: If the database does not exist.

        Example:
            >>> db = await AshenDB.get_db("test")
            >>> db
            <Database: test>
        """
        db_name = str(db_name)
        for name in await aios.listdir(cls.base):
            if name == db_name:
                return Database(cls.base + name + "/")
        raise NotFound(f"Database '{db_name}' does not exist.")

    @classmethod
    async def get_dbs(cls, db_names: list[str] = None) -> list[Database]:
        """Get multiple databases from list of database names.

        If no list is provided or the list is empty, all databases are returned.

        Args:
            db_names: The names of the databases.

        Raises:
            InvalidArgumentType: If the argument is not a list.
            NotFound: If one of the databases does not exist.

        Example:
            >>> dbs = await AshenDB.get_dbs(["test", "test2"])
            >>> dbs
            [<Database: test>, <Database: test2>]

            >>> dbs = await AshenDB.get_dbs()
            >>> dbs
            [<Database: test>, <Database: test2>, <Database: test3>]
        """
        final = []
        if db_names is None or len(db_names) == 0:
            for name in await aios.listdir(cls.base):
                final.append(Database(cls.base + name + "/"))
        elif isinstance(db_names, list) and len(db_names) > 0:
            for name in db_names:
                final.append(await cls.get_db(name))
        else:
            raise InvalidArgumentType(f"Expected list, got {type(db_names)}.")

        return final

    @classmethod
    async def iterate_dbs(cls, db_names: list[str]) -> Generator[Database, None, None]:
        """Iterate over multiple databases from list of database names.

        Args:
            db_names: The names of the databases.

        Raises:
            InvalidArgumentType: If the argument is not a list.
            NotFound: If one of the databases does not exist.

        Example:
            >>> async for db in AshenDB.iterate_dbs(["test", "test2"]):
            ...     print(db)
            <Database: test>
            <Database: test2>

            >>> dbs = [db async for db in AshenDB.iterate_dbs()]
            >>> dbs
            [<Database: test>, <Database: test2>, <Database: test3>]
        """
        if db_names is None or len(db_names) == 0:
            for name in await aios.listdir(cls.base):
                yield Database(cls.base + name + "/")
        elif isinstance(db_names, list) and len(db_names) > 0:
            for name in db_names:
                yield await cls.get_db(name)
        else:
            raise InvalidArgumentType(f"Expected list, got {type(db_names)}.")

    @classmethod
    async def create_db(cls, db_name: str or int) -> Database:
        """Creates a single database.

        Args:
            db_name: The name of the database.

        Raises:
            AlreadyExists: If the database already exists.

        Example:
            >>> db = await AshenDB.create_db("test")
            >>> db
            <Database: test>
        """
        db_name = str(db_name)
        path = cls.base + db_name
        if await aios.path.exists(path):
            raise AlreadyExists(f"Database '{db_name}' already exists.")
        else:
            await aios.mkdir(path)
            return Database(path + "/")

    @classmethod
    async def create_dbs(cls, db_names: list[str]) -> list[Database]:
        """Create multiple databases.

        Args:
            db_names: The names of the databases.

        Raises:
            AlreadyExists: If one of the databases already exists.
            InvalidArgumentType: If the argument is not a list or Empty or No Argument was passed.

        Example:
            >>> dbs = await AshenDB.create_dbs(["test", "test2"])
            >>> dbs
            [<Database: test>, <Database: test2>]
        """
        final = []
        if db_names is None or len(db_names) == 0 or not isinstance(db_names, list):
            raise InvalidArgumentType(f"Expected list, got {type(db_names)}.")

        for name in db_names:
            final.append(await cls.create_one(name))

        return final

    @classmethod
    async def iterate_create_dbs(
        cls, db_names: list[str]
    ) -> Generator[Database, None, None]:
        """Iterate over multiple databases from list of database names.

        Args:
            db_names: The names of the databases.

        Raises:
            AlreadyExists: If one of the databases already exists.
            InvalidArgumentType: If the argument is not a list or Empty or No Argument was passed.

        Example:
            >>> async for db in AshenDB.iterate_create_dbs(["test", "test2"]):
            ...     print(db)
            <Database: test>
            <Database: test2>

            >>> dbs = [db async for db in AshenDB.iterate_create_dbs()]
            >>> dbs
            [<Database: test>, <Database: test2>, <Database: test3>]
        """
        if db_names is None or len(db_names) == 0 or not isinstance(db_names, list):
            raise InvalidArgumentType(f"Expected list, got {type(db_names)}.")
        for name in db_names:
            yield await cls.create_one(name)

    @classmethod
    async def del_db(cls, db_name: str) -> None:
        """Delete a single database.

        Args:
            db_name: The name of the database.

        Raises:
            NotFound: If the database does not exist.

        Example:
            >>> await AshenDB.delete_db("test")

        """
        path = cls.base + db_name
        if await aios.path.exists(path):
            await aios.removedirs(path)
            return
        else:
            raise Exception(f"Database '{db_name}' does not exist.")

    @classmethod
    async def del_dbs(cls, db_names: list[str]) -> None:
        """Delete multiple databases.

        If no list is provided or the list is empty, all databases are deleted.

        Args:
            db_names: The names of the databases.

        Raises:
            NotFound: If one of the databases does not exist.

        Example:
            >>> await AshenDB.delete_dbs(["test", "test2"])
        """
        if db_names is None:
            for name in await aios.listdir(cls.base):
                await cls.delete_db(name)
            return
        for name in db_names:
            await cls.delete_db(name)
        return



import aiofiles, json, urllib, httpx, subprocess
import aiofiles.os as aios
from typing import Generator
from uuid import uuid4

from .helper import match_data, update_data
from .document import Document
from .exception import *


def gen_id() -> str:
    """Generate a random id.

    Returns:
        A random id.
    """
    return str(uuid4())


class Collection:
    def __init__(self, path):
        self.path = path

    async def w3m(self) -> None:
        """Show the collection in w3m.

        Note:
            Your system must have w3m installed.
        """

        json_data = []
        for file in await aios.scandir(self.path):
            async with aiofiles.open(file.path, "r") as f:
                json_data.append(json.loads(await f.read()))

        safe_data = urllib.parse.quote(str(json_data).replace("'", '"'))
        safe_data = safe_data.replace("%20", "+")
        data = "jsonData=" + safe_data

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url="https://tools.atatus.com/tools/json-to-html",
                data=data,
                headers={
                    "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8"
                },
            )
            response.raise_for_status()

        subprocess.run(
            ["w3m", "-dump", "-T", "text/html", "-cols", "1000"],
            input=response.text.replace(
                "<table>", '<table  border="1" cellpadding="10">'
            ),
            encoding="utf-8",
        )
        return None

    async def get_doc(self, *, id: str or int = None, query: dict = None) -> Document:
        """Get a single document.

        You can pass either an id or a query. If both are passed then the id will be used.

        Args:
            id: The id of the document.
            query: A query to match the document.

        Raises:
            ValueError: If neither id nor query is passed.
            NotFound: If no document is found.

        Example:
            >>> doc = await coll.get_doc(1)
            >>> doc
            <Document: 1>

            >>> doc = await coll.get_doc(query={"name": "test"})
            >>> print(doc)
            {"name": "test"}
        """
        id = str(id)
        try:
            # check if the "{self.path}/{id}.json" file exists
            path = f"{self.path}/{id}.json"
            if not await aios.path.exists(path):
                raise NotFound("No document found")
            doc = Document(path)
            await doc.__ainit__()
            return doc
        except NotFound:
            for file in await aios.scandir(self.path):
                async with aiofiles.open(file.path, "r") as f:
                    data = json.loads(await f.read())
                    if await match_data(data, query):
                        doc = Document(file.path)
                        await doc.__ainit__()
                        return doc
            raise NotFound("No document found")
        except Exception as e:
            print(e)
            raise ValueError("Either id or query must be provided")

    async def get_docs(
        self, ids: list[str or int] = None, query: dict = None
    ) -> list[Document]:
        """Get multiple documents.

        You can pass either a list of ids or a query. If both are passed then the ids will be used.

        Args:
            ids: The ids of the documents.
            query: A query to match the documents.

        Raises:
            ValueError: If neither ids nor query is passed.
            NotFound: If no documents are found.

        Example:
            >>> docs = await coll.get_docs([1, 2, 3])
            >>> docs
            [<Document: 1>, <Document: 2>, <Document: 3>]

            >>> docs = await coll.get_docs(query={"name": "test"})
            >>> docs
            [<Document: 1>, <Document: 2>, <Document: 3>]
        """
        final = []
        if ids:
            for id in ids:
                final.append(await self.get_doc(id=id))
            return final
        elif query:
            for file in await aios.scandir(self.path):
                async with aiofiles.open(file.path, "r") as f:
                    data = json.loads(await f.read())
                    if await match_data(data, query):
                        doc = Document(file.path)
                        await doc.__ainit__()
                        final.append(doc)
            if len(final) == 0 or len(final[0]) == 0:
                raise NotFound("No documents found")
            return final
        else:
            raise ValueError("Either ids or query must be provided")

    async def iterate_docs(
        self, *, ids: list[str or int] = None, query: dict = None
    ) -> Generator[Document, None, None]:
        """Iterate over multiple documents.

        You can pass either a list of ids or a query. If both are passed then the ids will be used.

        Args:
            ids: The ids of the documents.
            query: A query to match the documents.

        Raises:
            ValueError: If neither ids nor query is passed.
            NotFound: If no documents are found.

        Example:
            >>> async for doc in coll.iterate_docs([1, 2, 3]):
            ...     print(doc)
            {"name": "test1"}
            {"name": "test2"}
            {"name": "test3"}

            >>> async for doc in coll.iterate_docs(query={"name": "test"}):
            ...     print(doc)
            {"name": "test1"}
            {"name": "test2"}
            {"name": "test3"}

            >>> docs = [doc async for doc in coll.iterate_docs([1, 2, 3])]
            >>> docs
            [<Document: 1>, <Document: 2>, <Document: 3>]
        """
        if ids:
            for id in ids:
                yield await self.get_doc(id=id)
        elif query:
            for file in await aios.scandir(self.path):
                async with aiofiles.open(file.path, "r") as f:
                    data = json.loads(await f.read())
                    if await match_data(data, query):
                        doc = Document(file.path)
                        await doc.__ainit__()
                        yield doc
        else:
            for file in await aios.scandir(self.path):
                async with aiofiles.open(file.path, "r") as f:
                    data = json.loads(await f.read())
                    doc = Document(file.path)
                    await doc.__ainit__()
                    yield doc

    async def create_doc(self, data: dict) -> Document:
        """Create a document.

        Args:
            data: The data to be stored in the document.

        Raises:
            TypeError: If data is not a dict.

        Example:
            >>> doc = await coll.create_doc({"name": "test"})
            >>> doc
            <Document: 1>
        """
        if not isinstance(data, dict):
            raise TypeError("Data must be a dict")

        try:
            id = str(data["_id"])
        except KeyError:
            id = gen_id()
            data["_id"] = id
        # Check if already exist
        if await aios.path.exists(f"{self.path}/{id}.json"):
            raise AlreadyExists("Document already exist")
        async with aiofiles.open(f"{self.path}/{id}.json", "w") as f:
            await f.write(json.dumps(data))
        document = Document(f"{self.path}/{id}.json")
        await document.__ainit__()
        return document

    async def create_docs(self, datas: list[dict]) -> list[Document]:
        """Create multiple documents.

        Args:
            datas: The data to be stored in the documents.

        Raises:
            TypeError: If datas is not a list.

        Example:
            >>> docs = await coll.create_docs([{"name": "test1"}, {"name": "test2"}])
            >>> docs
            [<Document: 1>, <Document: 2>]
        """
        if not isinstance(datas, list):
            raise TypeError("Datas must be a list")

        final = []
        for data in datas:
            final.append(await self.create_doc(data))
        return final

    async def del_doc(self, id: str or int = None, query: dict = None) -> None:
        """Delete a document.

        You can pass either an id or a query. If both are passed then the id will be used.

        Args:
            id: The id of the document.
            query: A query to match the document.

        Raises:
            ValueError: If neither id nor query is passed.
            NotFound: If no document is found.

        Example:
            >>> await coll.delete_doc(1)
        """
        if id:
            doc = await self.get_doc(id)
            await doc.delete()
        elif query:
            doc = await self.get_doc(query=query)
            await doc.delete()
        else:
            raise ValueError("Either id or query must be provided")

    async def del_docs(self, ids: list[str or int] = None, query: dict = None) -> None:
        """Delete multiple documents.

        You can pass either a list of ids or a query. If both are passed then the ids will be used.

        Args:
            ids: The ids of the documents.
            query: A query to match the documents.

        Raises:
            ValueError: If neither ids nor query is passed.
            NotFound: If no documents are found.

        Example:
            >>> await coll.delete_docs([1, 2, 3])
        """
        if ids:
            for id in ids:
                await self.delete_doc(id)
        elif query:
            for doc in await self.get_docs(query=query):
                await doc.delete()
        else:
            raise ValueError("Either ids or query must be provided")

    async def update_doc(
        self, id: str or int = None, query: dict = None, data: dict = None
    ) -> Document:
        """Update a document.

        You can pass either an id or a query. If both are passed then the id will be tried first, if it fails then the query will be used.

        Note:
            Not all update operators are supported. See helper_module for more info.
        Args:
            id: The id of the document.
            query: A query to match the document.
            data: The data to update the document with.

        Raises:
            ValueError: If neither id nor query is passed.
            NotFound: If no document is found.

        Example:
            >>> old_doc = await coll.update_doc(query={"name": "test"}}
            >>> old_doc
            {"name": "test"}
            >>> await coll.update_doc(query={"name": "test"}, data={"$set":{"name": "test2"})}
            >>> new_doc = await coll.get_doc(query={"name": "test2"})
            >>> new_doc
            {"name": "test2"}
        """
        doc = await self.get_doc(id=id, query=query)
        new_doc = await update_data(doc, data)
        return new_doc

    async def count_docs(self) -> int:
        """Count the number of documents in the collection.

        Example:
            >>> await coll.count_docs()
            3
        """
        count = 0
        for _ in await aios.scandir(self.path):
            count += 1
        return count
